<?php

declare(strict_types=1);

namespace Drupal\bulk_edit_terms\Form;

use Drupal\bulk_edit_terms\UpdateAction;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Form\ConfirmFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\TempStore\PrivateTempStoreFactory;
use Drupal\Core\Url;
use Drupal\node\NodeInterface;
use Drupal\taxonomy\TermStorageInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Response;

/**
 * Provides a form for bulk editing term assignments on nodes.
 */
class NodeSelectTerms extends ConfirmFormBase {

  /**
   * The array of nodes to edit terms from.
   *
   * @var array
   */
  protected array $nodes = [];

  public function __construct(
    protected PrivateTempStoreFactory $tempStoreFactory,
    protected EntityTypeManager $entityTypeManager,
    protected AccountInterface $currentUser,
    protected TimeInterface $time,
  ) {}

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): static {
    return new static(
      $container->get('tempstore.private'),
      $container->get('entity_type.manager'),
      $container->get('current_user'),
      $container->get('datetime.time'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'node_select_terms';
  }

  /**
   * {@inheritdoc}
   */
  public function getQuestion(): TranslatableMarkup {
    return $this->t('Which term reference fields do you want to update?');
  }

  /**
   * {@inheritdoc}
   */
  public function getCancelUrl(): Url {
    return new Url('system.admin_content');
  }

  /**
   * {@inheritdoc}
   */
  public function getConfirmText(): TranslatableMarkup {
    return $this->t('Update terms');
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state): array|Response {
    $node_ids = $this->tempStoreFactory->get('node_edit_terms')->get($this->currentUser->id());
    if (empty($node_ids)) {
      return new RedirectResponse($this->getCancelUrl()->setAbsolute()->toString());
    }

    $this->nodes = $this->entityTypeManager->getStorage('node')->loadMultiple($node_ids);

    $term_storage = $this->entityTypeManager->getStorage('taxonomy_term');
    assert($term_storage instanceof TermStorageInterface);

    // Gather all taxonomy fields from the nodes and offer the options available
    // to select.
    $field_names = [];
    $terms_fields_present = FALSE;
    foreach ($this->nodes as $node) {
      assert($node instanceof NodeInterface);
      $fields = $node->getFieldDefinitions();
      foreach ($fields as $name => $field) {
        if (!isset($form[$name])) {
          assert($field instanceof FieldDefinitionInterface);
          if ($field->getType() === 'entity_reference' && str_starts_with($name, 'field_')) {
            $field_info = $field->getFieldStorageDefinition();
            $field_settings = $field->getSettings();
            $target_bundles = $field_settings['handler_settings']['target_bundles'];

            if ($field_settings['target_type'] !== 'taxonomy_term') {
              continue;
            }

            $field_names[] = $name;

            $terms_fields_present = TRUE;

            $auto_create = (!empty($field->getSettings()['handler_settings']['auto_create'])) ?
              $field->getSettings()['handler_settings']['auto_create'] :
              FALSE;

            $single_value_only = $field_info->getCardinality() === 1;

            $auto_create_bundle = ($auto_create && !empty($field_settings['handler_settings']['auto_create_bundle'])) ?
              $field_settings['handler_settings']['auto_create_bundle'] :
              NULL;

            $form["{$name}_group"] = [
              '#type' => 'fieldset',
              '#title' => $field->getLabel(),
            ];

            if ($single_value_only && $target_bundles && !$auto_create) {
              $form["{$name}_group"]["{$name}_action"] = [
                '#type' => 'radios',
                '#title' => $this->t('Action'),
                '#options' => [
                  UpdateAction::Replace->value => $this->t('Replace with selected term'),
                  UpdateAction::Clear->value => $this->t('Clear existing value'),
                ],
              ];

              // Search terms and create select element.
              $terms = [];
              foreach ($target_bundles as $target_bundle) {
                $tree = $term_storage->loadTree($target_bundle);
                if (!empty($tree)) {
                  foreach ($tree as $item) {
                    $terms[$item->tid] = str_repeat('-', $item->depth) . $item->name;
                  }
                }
              }
              $options = ['' => '- Select -'];
              $options += $terms;
              $form["{$name}_group"][$name] = [
                '#type' => 'select',
                '#title' => $this->t('Term'),
                '#options' => $options,
                '#states' => [
                  'invisible' => [
                    'input[name="' . $name . '_action"]' => ['value' => UpdateAction::Clear->value],
                  ],
                ],
              ];
            }
            else {
              $form["{$name}_group"]["{$name}_action"] = [
                '#type' => 'radios',
                '#title' => $this->t('Action'),
                '#options' => [
                  UpdateAction::Replace->value => $this->t('Replace with selected term(s)'),
                  UpdateAction::Clear->value => $this->t('Clear existing value'),
                  UpdateAction::Append->value => $this->t('Add selected term(s)'),
                  UpdateAction::Remove->value => $this->t('Remove selected term(s)'),
                ],
              ];

              $form["{$name}_group"][$name] = [
                '#type' => 'entity_autocomplete',
                '#target_type' => $field_settings['target_type'],
                '#title' => $this->t('Term(s)'),
                '#default_value' => FALSE,
                '#tags' => TRUE,
                '#selection_settings' => [
                  'target_bundles' => $target_bundles,
                ],
                '#states' => [
                  'invisible' => [
                    'input[name="' . $name . '_action"]' => ['value' => UpdateAction::Clear->value],
                  ],
                ],
              ];

              if ($auto_create && $auto_create_bundle) {
                $form[$name]['#autocreate'] = [
                  'bundle' => $auto_create_bundle,
                ];
              }
            }
          }
        }
      }
    }

    // Back out now if we don't have any term reference fields that can be
    // modified.
    if (!$terms_fields_present) {
      $form['warning'] = [
        '#prefix' => '<p>',
        '#suffix' => '</p>',
        '#markup' => $this->t('No term reference fields were found in the selected nodes.'),
      ];
      return $form;
    }

    $form['intro'] = [
      '#prefix' => '<p>',
      '#suffix' => '</p>',
      '#markup' => $this->t('The following fields appear in at least one of the selected content items.<br/>Select actions and values for each field and these will be applied only to the content items that have that field.'),
      '#weight' => -1,
    ];

    $form['create_revision'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Create new revision for each update'),
      '#default_value' => TRUE,
    ];

    $form['revision_log_message'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Revision log message'),
      '#rows' => 4,
      '#states' => [
        'visible' => [
          'input[name="create_revision"]' => ['checked' => TRUE],
        ],
      ],
    ];

    // Store list of field names we presented on the form so we can access
    // them in the form submit.
    $form_state->set('field_names', $field_names);

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state): void {
    foreach ($form_state->get('field_names') as $fieldName) {
      $field_value = $form_state->getValue($fieldName);
      $field_action = $form_state->getValue($fieldName . '_action');

      if (empty($field_action) && !empty($field_value)) {
        $form_state->setErrorByName($fieldName, $this->t('Select an action for each field you provided a value for.'));
      }
      if (empty($field_value) && !empty($field_action) && $field_action !== UpdateAction::Clear->value) {
        $form_state->setErrorByName($fieldName, $this->t('You must provide a value for each field.'));
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    if ($form_state->getValue('confirm') && !empty($this->nodes)) {
      $updated_nodes_count = 0;
      foreach ($this->nodes as $node) {
        assert($node instanceof NodeInterface);
        $changes_made_to_node = FALSE;
        foreach ($form_state->get('field_names') as $fieldName) {
          // Skip if this node doesn't have this field.
          $field = $node->getFieldDefinition($fieldName);
          if (!$field instanceof FieldDefinitionInterface) {
            continue;
          }

          // Skip if there's no action to take for this field.
          $field_value = $form_state->getValue($fieldName);
          $field_action = $form_state->getValue($fieldName . '_action');
          if (empty($field_action)) {
            continue;
          }

          $changes_made_to_field_value = FALSE;
          $value_to_set = NULL;
          $field_action = UpdateAction::from($field_action);
          $existing_values = $node->get($fieldName)->getValue();
          switch ($field_action) {
            case UpdateAction::Clear:
              if (!$node->get($fieldName)->isEmpty()) {
                $value_to_set = [];
                $changes_made_to_field_value = TRUE;
              }
              break;

            case UpdateAction::Replace:
              // Only take action if there was no existing value or the existing
              // value doesn't match the new value.
              if ($node->get($fieldName)->isEmpty() || $node->get($fieldName)->first()->getValue()['target_id'] != $field_value) {
                $value_to_set = $field_value;
                $changes_made_to_field_value = TRUE;
              }
              break;

            case UpdateAction::Remove:
              $tids_to_remove = array_map(fn($val) => $val['target_id'], $field_value);
              $value_to_set = array_filter($existing_values, fn ($value) => !in_array($value['target_id'], $tids_to_remove));
              if ($value_to_set !== $existing_values) {
                $changes_made_to_field_value = TRUE;
              }
              break;

            case UpdateAction::Append:
              $existing_values = $node->get($fieldName)->getValue();
              // Remove duplicates from $value.
              $field_value = array_udiff($field_value, $existing_values, function ($a, $b) {
                return $a['target_id'] <=> $b['target_id'];
              });
              $value_to_set = array_merge($field_value, $existing_values);
              if ($value_to_set !== $existing_values) {
                $changes_made_to_field_value = TRUE;
              }
              break;
          }

          if ($changes_made_to_field_value) {
            $node->set($fieldName, $value_to_set);
            $changes_made_to_node = TRUE;
          }
        }

        if ($changes_made_to_node) {
          if ($form_state->getValue('create_revision')) {
            $msg = $form_state->getValue('revision_log_message');
            if (!empty($msg)) {
              $node->set('revision_log', $msg);
            }
            $node->setNewRevision();
            $node->setRevisionUserId($this->currentUser->id());
            $node->setRevisionCreationTime($this->time->getCurrentTime());
          }
          $node->save();
          $updated_nodes_count++;
        }
      }

      $this->messenger()->addStatus($this->formatPlural($updated_nodes_count, '1 node was updated.', '@count nodes were updated.'));
    }
  }

}
